# Copyright (C) 2025  Johannes Sasongko <johannes sasongko org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.


import os.path
import shelve
import tempfile

try:
    import berkeleydb
except ImportError:
    import bsddb3 as berkeleydb

from xl import common


def test_migration_to_sqlite():
    with tempfile.TemporaryDirectory(prefix="exaile-") as tmpdir:
        dbpath = os.path.join(tmpdir, "music.db")

        with shelve.BsdDbShelf(
            berkeleydb.hashopen(dbpath, 'c'), protocol=common.PICKLE_PROTOCOL
        ) as db:
            db['key'] = ({'tag': 'value'}, 2, {})

        with common.open_shelf(dbpath) as db:
            assert list(db.items()) == [('key', ({'tag': 'value'}, 2, {}))]
